#include "MazeGenerator.h"


const int MazeGenerator::DIRECTION_X[9] = { 0, 0,+1, 0, 0, 0, 0, 0,-1 };
const int MazeGenerator::DIRECTION_Y[9] = { 0,-1, 0, 0,+1, 0, 0, 0, 0 };


const int MazeGenerator::MASK[9] = {
	0,
	SOUTH | SOUTH << 4,
	WEST | WEST << 4,
	0,
	NORTH | NORTH << 4,
	0,
	0,
	0,
	EAST | EAST << 4
};

MazeGenerator::MazeGenerator()
{
}


MazeGenerator::~MazeGenerator()
{
}

std::string MazeGenerator::generate(int size, std::string filename)
{

	std::string info;
	
	if (size <= 0) {

		size = 10;

		info = "Number of cells was too low. A maze of 10x10 cells has been generated.";

	}
	else if (size > 100)
	{
		size = 100;

		info = "Number of cells was too high. A maze of 100x100 cells has been generated.";

	}
	else {

		info = "Maze of given number of cells has been generated.";

	}
	
	numberOfCells_ = size;

	cellSize_ = 10;

	imageSize_ = (numberOfCells_ + 1) * cellSize_;

	maze_ = new unsigned char[numberOfCells_* numberOfCells_];

	srand(time(NULL));

	std::fill(maze_, maze_ + numberOfCells_ * numberOfCells_, 0);

	idX_ = getRandomInt();
	idY_ = getRandomInt();

	generateMaze();

	BMP image;

	image.SetSize(imageSize_ + cellSize_ - 1, imageSize_ + cellSize_ - 1);
	image.SetBitDepth(8);

	setColor(backgroundColor_, 255, 255, 255);
	setColor(mazeColor_, 0, 0, 0);

	drawMaze(image);

	image.WriteToFile(filename.c_str());

	return info;

}


void MazeGenerator::drawBackground(BMP& image)
{
	for (int i = 0; i < image.TellWidth(); i++)
	{
		for (int j = 0; j < image.TellHeight(); j++)
		{
			image.SetPixel(i, j, backgroundColor_);
		}
	}
}


void MazeGenerator::setColor(RGBApixel& pixel, const int& r, const int& g, const int& b)
{
	pixel.Red = r;
	pixel.Green = g;
	pixel.Blue = b;	
}


bool MazeGenerator::isDirectionValid(const Direction& direction)
{
	int NewX = idX_ + DIRECTION_X[direction];
	int NewY = idY_ + DIRECTION_Y[direction];

	if (!direction || NewX < 0 || NewY < 0 || NewX >= numberOfCells_ || NewY >= numberOfCells_) return false;

	return !maze_[NewX + numberOfCells_ * NewY];
}


MazeGenerator::Direction MazeGenerator::getDirection()
{
	Direction direction = Direction(1 << getRandomInt4());

	while (true)
	{
		for (int x = 0; x < 4; x++)
		{
			if (isDirectionValid(direction)) { return Direction(direction); }

			direction = Direction(direction << 1);

			if (direction > WEST) { direction = NORTH; }
		}

		direction = Direction((maze_[getCellId()] & 0xf0) >> 4);

		if (!direction) return INVALID;

		idX_ += DIRECTION_X[direction];
		idY_ += DIRECTION_Y[direction];

		direction = Direction(1 << getRandomInt4());
	}
}


void MazeGenerator::generateMaze()
{

	for (Direction direction = getDirection(); direction != INVALID; direction = getDirection())
	{

		maze_[getCellId()] |= direction;

		idX_ += DIRECTION_X[direction];
		idY_ += DIRECTION_Y[direction];

		maze_[getCellId()] = MASK[direction];
	}


}



void MazeGenerator::drawLine(BMP& maze, const int& x1, const int& y1, const int& x2, const int& y2)
{
	if (x1 == x2)
	{
		for (int y = y1; y < y2; y++)
		{
			if (x1 >= imageSize_ || y >= imageSize_) continue;
			maze.SetPixel(x1 + cellSize_ - 1, y + cellSize_ - 1, mazeColor_);
			
		}
	}

	if (y1 == y2)
	{
		for (int x = x1; x < x2; x++)
		{
			if (y1 >= imageSize_ || x >= imageSize_) continue;
			maze.SetPixel(x + cellSize_ - 1, y1 + cellSize_ - 1, mazeColor_);
			
		}
	}
}


void MazeGenerator::drawMaze(BMP& image)
{
	drawBackground(image);

	for (int y = 0; y < numberOfCells_; y++)
	{
		for (int x = 0; x < numberOfCells_; x++)
		{
			char v = maze_[y * numberOfCells_ + x];

			int nx = x * cellSize_;
			int ny = y * cellSize_;

			if (!(v & NORTH)) drawLine(image, nx, ny, nx + cellSize_ + 1, ny);
			if (!(v & EAST)) drawLine(image, nx + cellSize_, ny, nx + cellSize_, ny + cellSize_ + 1);
			if (!(v & SOUTH)) drawLine(image, nx, ny + cellSize_, nx + cellSize_ + 1, ny + cellSize_);
			if (!(v & WEST)) drawLine(image, nx, ny, nx, ny + cellSize_ + 1);
		}
	}
}

